"""
A permutation-with-repetition-based encoding based on games.

A point in the search space is a permutation (potentially with repetitions)
that can be translated to a :class:`~moptipyapps.ttp.game_plan.GamePlan`.
Each value `v` in the permutation represents a game to be played by two of
the `n` teams. There are `n(n-1)` possible games between `n` teams,
distinguishing home and away teams. Given a value `v` from `0..n(n-1)-1`,
we can get the zero-based index of the home team as
`home_idx = (game // (n - 1)) % n`. The away index is computed in two steps,
first we set `away_idx = game % (n - 1)` and if `away_idx >= home_idx`, we
do `away_idx = away_idx + 1`. (Because a team can never play against itself,
the situation that `home_idx == away_idx` does not need to be represented, so
we can "skip" over this possible value by doing the `away_idx = away_idy + 1`
and thus get a more "compact" numeric range for the permutation elements.)

A game schedule for any round-robin tournament with any given number of rounds
can then be represented as permutation (potentially with repetitions) of these
game values. In the decoding procedure, it is processed from beginning to end
each game is then placed into the earliest slot not already occupied by
another game. In other words, it is placed at the earliest day at which both
involved teams do not yet have other games. If no such slot is available, this
game is not placed at all. In this case, there will be some zeros in the game
plan after the encoding. No other constraint is considered at this stage.

In other words, this encoding may produce game plans that violate constraints.
It does not care about the streak length constraints.
It does not ensure that each team always has a game.
Therefore, it should only be used in conjunction with objective functions that
force the search towards feasible solutions, such as the
:mod:`~moptipyapps.ttp.errors` objective.

In :mod:`~moptipyapps.ttp.game_encoding_2`, we present a modified version of
this encoding procedure that tries to reduce the number of zeros, i.e.,
"byes," in the resulting schedule.
"""


from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.encoding import Encoding
from moptipy.spaces.permutations import Permutations
from pycommons.types import check_int_range

from moptipyapps.ttp.instance import Instance


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def game_to_id(home: int | np.ndarray, away: int | np.ndarray, n: int) -> int:
    """
    Encode a game to a game ID.

    :param home: the home team
    :param away: the away team
    :param n: the total number of teams
    :return: the game ID

    >>> game_to_id(0, 1, 2)
    0
    >>> game_to_id(1, 0, 2)
    1

    >>> game_to_id(0, 1, 4)
    0
    >>> game_to_id(0, 2, 4)
    1
    >>> game_to_id(0, 3, 4)
    2
    >>> game_to_id(1, 0, 4)
    3
    >>> game_to_id(1, 2, 4)
    4
    >>> game_to_id(1, 3, 4)
    5
    >>> game_to_id(2, 0, 4)
    6
    >>> game_to_id(2, 1, 4)
    7
    >>> game_to_id(2, 3, 4)
    8
    >>> game_to_id(3, 0, 4)
    9
    >>> game_to_id(3, 1, 4)
    10
    >>> game_to_id(3, 2, 4)
    11
    >>> game_to_id(5, 4, 10)
    49
    """
    a = int(home)
    b = int(away)
    if b > a:
        b -= 1
    return (a * (n - 1)) + b


def search_space_for_n_and_rounds(n: int, rounds: int) -> Permutations:
    """
    Create a proper search space for the given number of teams and rounds.

    If the instance prescribes a double-round robin tournament, then this
    is just the :meth:`~moptipy.spaces.permutations.Permutations.standard`
    permutations set. Otherwise, it will be a permutation where some
    elements are omitted (for
    :attr:`~moptipyapps.ttp.instance.Instance.rounds` == 1) or duplicated
    (if :attr:`~moptipyapps.ttp.instance.Instance.rounds` > 2).

    If an odd number of rounds is played, then it is not possible that all
    teams have the same number of games at home and away. Then, the
    permutation is generated such that, if the highest numbers of games at
    home for any team is `k`, no other team has less than `k - 1` games at
    home. If the number of rounds is even, then all teams will have the
    same number of home and away games, that is, the number of teams
    divided by two and multiplied by the number of rounds.

    :param n: the number of teams
    :param rounds: the number of rounds
    :return: the search space

    >>> ",".join(map(str, search_space_for_n_and_rounds(2, 2).blueprint))
    '0,1'
    >>> ",".join(map(str, search_space_for_n_and_rounds(2, 3).blueprint))
    '0,1,1'
    >>> ",".join(map(str, search_space_for_n_and_rounds(2, 4).blueprint))
    '0,0,1,1'
    >>> ",".join(map(str, search_space_for_n_and_rounds(2, 5).blueprint))
    '0,0,1,1,1'
    >>> ",".join(map(str, search_space_for_n_and_rounds(3, 1).blueprint))
    '1,2,5'
    >>> ",".join(map(str, search_space_for_n_and_rounds(3, 2).blueprint))
    '0,1,2,3,4,5'
    >>> ",".join(map(str, search_space_for_n_and_rounds(3, 3).blueprint))
    '0,1,1,2,2,3,4,5,5'
    >>> ",".join(map(str, search_space_for_n_and_rounds(4, 1).blueprint))
    '1,2,3,7,8,10'
    >>> ",".join(map(str, search_space_for_n_and_rounds(4, 2).blueprint))
    '0,1,2,3,4,5,6,7,8,9,10,11'
    >>> ",".join(map(str, search_space_for_n_and_rounds(4, 3).blueprint))
    '0,1,1,2,2,3,3,4,5,6,7,7,8,8,9,10,10,11'
    >>> ",".join(map(str, search_space_for_n_and_rounds(4, 4).blueprint))
    '0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11'
    >>> ",".join(map(str, search_space_for_n_and_rounds(4, 5).blueprint))
    '0,0,1,1,1,2,2,2,3,3,3,4,4,5,5,6,6,7,7,7,8,8,8,9,9,10,10,10,11,11'
    >>> ",".join(map(str, search_space_for_n_and_rounds(5, 1).blueprint))
    '1,2,4,7,9,10,13,15,16,18'
    >>> ",".join(map(str, search_space_for_n_and_rounds(5, 2).blueprint))
    '0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19'
    >>> ",".join(map(str, search_space_for_n_and_rounds(5, 3).blueprint))
    '0,1,1,2,2,3,4,4,5,6,7,7,8,9,9,10,10,11,12,13,13,14,15,15,16,16,17,18,\
18,19'
    """
    check_int_range(n, "n", 2, 100000)
    check_int_range(n, "rounds", 1, 100000)
    games: Final[list[int]] = []
    order: bool = False
    for r in range(rounds):
        # If we have an odd round of games, the very last round needs
        # to be treated differently to ensure that the home-away games
        # distribution is fair.
        normal: bool = (r < (rounds - 1)) or ((rounds % 2) == 0)
        for i in range(n):  # for each city
            for j in range(i):  # for each other city
                order = ((r % 2) == 0) if normal else (not order)
                games.append(game_to_id(
                    i if order else j, j if order else i, n))
    games.sort()
    return Permutations(games)  # create permutations with repetition


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def re_encode(x: np.ndarray, y: np.ndarray) -> None:
    """
    Re-encode a game plan to a permutation.

    `x` must be a valid permutation. It is then transformed based on the game
    plan `y` to represent `y` in a straightforward manner. The general
    contract of this function is that:

    - `y` can be a game plan of any kind, with all sorts of game scheduling
       error except one: If team A plays team B in one slot, then team B must
       also play against team A in that slot and exactly one of them is the
       home team and one of them is the away team. Apart from that, there can
       be arbitrary byes or streak violations.
    - `x` must a valid game permutation, but can be entirely unrelated
       to `y`.
    - This function will then transform `x` such that `map_games(x, z)` will
      result in a game plan such that `z = y`.

    Due to the possibility of byes, it can happen that after
    `map_games(x1, z)` and `re_encode(x2, z)` it may be that `x1 != x2`.
    The reason is that the order of game IDs n `x2` for games that were not
    scheduled in `z` will be undefined.

    :param x: the permutation
    :param y: the game plan

    >>> from moptipyapps.ttp.instance import Instance
    >>> inst = Instance.from_resource("con14")
    >>> perm = np.array([70,114,54,51,98,31,49,46,148,169,174,151,155,110,
    ...     75,118,128,81,171,19,133,93,5,91,47,1,84,109,142,121,7,40,163,
    ...     61,20,96,165,177,160,39,73,37,101,108,119,65,2,117,164,106,60,
    ...     145,105,90,170,74,8,26,94,10,22,86,120,154,89,66,77,178,140,48,
    ...     0,52,159,25,176,135,63,57,76,161,68,124,162,29,55,80,24,45,85,
    ...     115,131,100,17,53,27,38,95,158,104,175,87,123,167,97,79,112,172,
    ...     127,136,12,44,50,102,69,144,143,113,4,181,88,166,150,59,141,72,
    ...     125,116,132,14,147,153,129,111,92,62,6,32,82,16,179,21,3,126,
    ...     134,122,149,146,36,107,34,103,42,41,18,35,78,28,137,43,33,71,99,
    ...     139,56,23,138,67,168,130,30,180,152,83,9,13,173,11,157,156,64,
    ...     15,58])
    >>> teams = 14
    >>> rounds = 2
    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ -8 -10  -5  14   3   7  -6   1  12   2 -13  -9  11  -4]
     [-14  -6   7  12  11   2  -3   9  -8  13  -5  -4 -10   1]
     [  7   8 -14   9 -10 -12  -1  -2  -4   5  13   6 -11   3]
     [ -5  11  -8  -7   1 -14   4   3 -12 -13  -2   9  10   6]
     [  3  -5  -1 -11   2  10  -9 -13   7  -6   4  14   8 -12]
     [  9  -3   2  10 -13  12   8  -7  -1  -4  14  -6   5 -11]
     [-10  -4  13   2 -11  -9  14  12   6   1   5  -8  -3  -7]
     [ -4   9 -10   1   7  -8  -5   6  -2   3 -14  13 -12  11]
     [ -6 -11 -12  -8  10   1  13   4 -14  -5   2   3  -7   9]
     [  4 -14 -13  -1  -9  11  10 -12   5  -7  -6   8   3   2]
     [ 10  -7   5   8  -3  14   2  -4 -13  -1  12 -11   9  -6]
     [ 12  14  -9 -10  13 -11  -8   7   3   4   6  -1  -5  -2]
     [ -3  -9   1  11  -8  13  12   5   2 -14  -4  -7  -6  10]
     [  2  -1  -7 -13  -6   5   3 -14  10  -9 -12  11   4   8]
     [-11 -12  14  -5   4 -13   9 -10  -7   8   1   2   6  -3]
     [ -9   3  -2  -6 -14   4 -13  11   1  12  -8 -10   7   5]
     [-12  13   8   5  -4 -10 -14  -3  11   6  -9   1  -2   7]
     [  8   6  10 -14 -12  -2  11  -1  13  -3  -7   5  -9   4]
     [ 14  -8 -11   6   9  -4 -10   2  -5   7   3 -13  12  -1]
     [  6   0   0  13  12  -1 -11  14 -10   9   7  -5  -4  -8]
     [ 11   5  12   7  -2   9  -4  13  -6  14  -1  -3  -8 -10]
     [  0  10  11 -12  -7   0   5  -9   8  -2  -3   4 -14  13]
     [  5 -13  -4   3  -1   8 -12  -6  14  11 -10   7   2  -9]
     [  0   7   0  -9   6  -5  -2 -11   4 -12   8  10  14 -13]
     [ -7  12   4  -3  14   0   1  10 -11  -8   9  -2   0  -5]
     [ -2   1   9   0   8  -7   6  -5  -3 -11  10 -14   0  12]]
    >>> from moptipyapps.ttp.game_encoding_2 import map_games_2
    >>> dest.fill(0)
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ -8 -10  -5  14   3   7  -6   1  12   2 -13  -9  11  -4]
     [-14  -6   7  12  11   2  -3   9  -8  13  -5  -4 -10   1]
     [  7   8 -14   9 -10 -12  -1  -2  -4   5  13   6 -11   3]
     [ -5  11  -8  -7   1 -14   4   3 -12 -13  -2   9  10   6]
     [  3  -5  -1 -11   2  10  -9 -13   7  -6   4  14   8 -12]
     [  9  -3   2  10 -13  12   8  -7  -1  -4  14  -6   5 -11]
     [-10  -4  13   2 -11  -9  14  12   6   1   5  -8  -3  -7]
     [ -4   9 -10   1   7  -8  -5   6  -2   3 -14  13 -12  11]
     [ -6 -11 -12  -8  10   1  13   4 -14  -5   2   3  -7   9]
     [  4 -14 -13  -1  -9  11  10 -12   5  -7  -6   8   3   2]
     [ 10  -7   5   8  -3  14   2  -4 -13  -1  12 -11   9  -6]
     [ 12  14  -9 -10  13 -11  -8   7   3   4   6  -1  -5  -2]
     [ -3  -9   1  11  -8  13  12   5   2 -14  -4  -7  -6  10]
     [  2  -1  -7 -13  -6   5   3 -14  10  -9 -12  11   4   8]
     [-11 -12  14  -5   4 -13   9 -10  -7   8   1   2   6  -3]
     [ -9   3  -2  -6 -14   4 -13  11   1  12  -8 -10   7   5]
     [-12  13   8   5  -4 -10 -14  -3  11   6  -9   1  -2   7]
     [  8   6  10 -14 -12  -2  11  -1  13  -3  -7   5  -9   4]
     [ 14  -8 -11   6   9  -4 -10   2  -5   7   3 -13  12  -1]
     [ -2   1  -6  13  12   3 -11  14 -10   9   7  -5  -4  -8]
     [ 11   5  12   7  -2   9  -4  13  -6  14  -1  -3  -8 -10]
     [  6  10  11 -12  -7  -1   5  -9   8  -2  -3   4 -14  13]
     [  5 -13  -4   3  -1   8 -12  -6  14  11 -10   7   2  -9]
     [-13   7   6  -9  14  -3  -2 -11   4 -12   8  10   1  -5]
     [ -7  12   4  -3   6  -5   1  10 -11  -8   9  -2  14 -13]
     [ 13   4   9  -2   8  -7   6  -5  -3 -11  10 -14  -1  12]]
    >>> re_encode(perm, dest)
    >>> print(perm)
    [ 51  54  70  91 114 118 166  31  49  61  66  98 128 169   5  19  46 121
     141 148 171  22  52  81  93 151 165 174   1  53  73 110 133 155 163   7
      27  47  75  84 142 160  37  40  90 101 109 117 134  20  39  57  96 119
     154 179  60  65  89  94 131 145 177   2  74  86 108 150 158 170   8  29
      45  77  79 140 164  10  25  63  97 106 120 135  26  48  76  88  95 105
     178   0  69  80 112 153 159 176  38  55  85 124 130 144 161  14  68 100
     104 127 162 173  24  32  42 113 122 143 175   6  17  34  87 115 147 172
      12  43  59  92 123 132 167  13  50  62  67 103 125 136   9  16  36  44
      72 102 129   4  21  35  82 111 146 181   3  41  71 116 126 149 157  18
      30  64 107 137 152 156  23  28  56  78  99 138 168  11  15  33  58  83
     139 180]
    >>> dest.fill(0)
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ -8 -10  -5  14   3   7  -6   1  12   2 -13  -9  11  -4]
     [-14  -6   7  12  11   2  -3   9  -8  13  -5  -4 -10   1]
     [  7   8 -14   9 -10 -12  -1  -2  -4   5  13   6 -11   3]
     [ -5  11  -8  -7   1 -14   4   3 -12 -13  -2   9  10   6]
     [  3  -5  -1 -11   2  10  -9 -13   7  -6   4  14   8 -12]
     [  9  -3   2  10 -13  12   8  -7  -1  -4  14  -6   5 -11]
     [-10  -4  13   2 -11  -9  14  12   6   1   5  -8  -3  -7]
     [ -4   9 -10   1   7  -8  -5   6  -2   3 -14  13 -12  11]
     [ -6 -11 -12  -8  10   1  13   4 -14  -5   2   3  -7   9]
     [  4 -14 -13  -1  -9  11  10 -12   5  -7  -6   8   3   2]
     [ 10  -7   5   8  -3  14   2  -4 -13  -1  12 -11   9  -6]
     [ 12  14  -9 -10  13 -11  -8   7   3   4   6  -1  -5  -2]
     [ -3  -9   1  11  -8  13  12   5   2 -14  -4  -7  -6  10]
     [  2  -1  -7 -13  -6   5   3 -14  10  -9 -12  11   4   8]
     [-11 -12  14  -5   4 -13   9 -10  -7   8   1   2   6  -3]
     [ -9   3  -2  -6 -14   4 -13  11   1  12  -8 -10   7   5]
     [-12  13   8   5  -4 -10 -14  -3  11   6  -9   1  -2   7]
     [  8   6  10 -14 -12  -2  11  -1  13  -3  -7   5  -9   4]
     [ 14  -8 -11   6   9  -4 -10   2  -5   7   3 -13  12  -1]
     [ -2   1  -6  13  12   3 -11  14 -10   9   7  -5  -4  -8]
     [ 11   5  12   7  -2   9  -4  13  -6  14  -1  -3  -8 -10]
     [  6  10  11 -12  -7  -1   5  -9   8  -2  -3   4 -14  13]
     [  5 -13  -4   3  -1   8 -12  -6  14  11 -10   7   2  -9]
     [-13   7   6  -9  14  -3  -2 -11   4 -12   8  10   1  -5]
     [ -7  12   4  -3   6  -5   1  10 -11  -8   9  -2  14 -13]
     [ 13   4   9  -2   8  -7   6  -5  -3 -11  10 -14  -1  12]]
    """
    days, n = y.shape  # the number of days and teams to be scheduled
    total: Final[int] = len(x)
    index: int = 0
    for day in range(days):
        for slot in range(n):
            other = int(y[day, slot])
            if other > 0:
                game_id = game_to_id(slot, other - 1, n)
                for swap in range(index, total):
                    if x[swap] == game_id:
                        x[swap] = x[index]
                        x[index] = game_id
                        break
                index += 1


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def map_games(x: np.ndarray, y: np.ndarray) -> None:
    """
    Translate a permutation of games to a game plan.

    This is a straightforward decoding that places the games into the map one
    by one. Each game is placed at the earliest slot in which it can be
    placed. If a game cannot be placed, it is ignored. This will lead to many
    errors, which can be counted via the :mod:`~moptipyapps.ttp.errors`
    objective.

    :param x: the source permutation
    :param y: the destination game plan

    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> teams = 2
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> print(perm)
    [0 1]
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1]
     [-2  1]]
    >>> perm[0] = 1
    >>> perm[1] = 0
    >>> re_encode(perm, dest)
    >>> print(perm)
    [0 1]
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1]
     [-2  1]]
    >>> teams = 4
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 4  3 -2 -1]
     [-2  1 -4  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    >>> re_encode(perm, dest)
    >>> print(perm)
    [ 0  8  1  5  2  4  3 11  6 10  7  9]
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 4  3 -2 -1]
     [-2  1 -4  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
     >>> from moptipyapps.ttp.instance import Instance
     >>> inst = Instance.from_resource("circ10")
     >>> perm = np.array([73,77,55,74,21,20,3,11,63,19,38,8,27,47,88,16,75,
     ...     45,89,36,24,80,17,40,0,32,53,82,31,13,12,66,71,6,87,84,2,60,61,
     ...     10,30,58,49,57,54,23,26,46,59,42,29,62,22,18,50,78,37,34,65,43,
     ...     48,85,86,83,56,41,5,81,52,15,51,9,4,76,7,69,67,33,79,72,35,25,
     ...     1,14,70,44,68,64,28,39])
    >>> teams = 10
    >>> rounds = 2
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ -8  -9   5   7  -3  10  -4   1   2  -6]
     [  5  -7   4  -3  -1  -9   2 -10   6   8]
     [ 10   4  -9  -2   6  -5   8  -7   3  -1]
     [ -4  -3   2   1  -7   8   5  -6 -10   9]
     [ -6   9  -5  -8   3   1 -10   4  -2   7]
     [ -5  10  -6  -9   1   3  -8   7   4  -2]
     [  2  -1   8   6   7  -4  -5  -3  10  -9]
     [  8 -10   6   5  -4  -3   9  -1  -7   2]
     [  4   6   7  -1   9  -2  -3  10  -5  -8]
     [ -7   5  -8 -10  -2   9   1   3  -6   4]
     [-10   3  -2  -7  -6   5   4  -9   8   1]
     [  0  -6  10   0   8   2  -9  -5   7  -3]
     [  9  -5  -4   3   2  -7   6   0  -1   0]
     [ -3   8   1   9   0   0  10  -2  -4  -7]
     [ -2   1   9   8 -10   7  -6  -4  -3   5]
     [  7  -8 -10  -6  -9   4  -1   2   5   3]
     [ -9  -4  -7   2  -8 -10   3   5   1   6]
     [  6   7   0  10   0  -1  -2   9  -8  -4]]
    >>> int(dest.shape[0] * dest.shape[1] - np.count_nonzero(dest))
    8
    >>> perm = np.array([21,32,53,63,73,3,20,55,77,88,8,11,40,60,74,19,27,51,
    ...     58,89,16,38,45,66,87,17,36,47,69,75,0,24,31,41,80,6,22,30,61,82,
    ...     2,13,23,43,71,12,52,54,65,84,10,49,57,79,81,1,28,44,68,78,7,26,
    ...     39,59,64,18,34,42,46,62,9,25,33,50,85,5,15,48,76,83,14,29,67,72,
    ...     86,4,35,37,56,70])
    >>> dest.fill(0)
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ -8  -9   5   7  -3  10  -4   1   2  -6]
     [  5  -7   4  -3  -1  -9   2 -10   6   8]
     [ 10   4  -9  -2   6  -5   8  -7   3  -1]
     [ -4  -3   2   1  -7   8   5  -6 -10   9]
     [ -6   9  -5  -8   3   1 -10   4  -2   7]
     [ -5  10  -6  -9   1   3  -8   7   4  -2]
     [  2  -1   8   6   7  -4  -5  -3  10  -9]
     [  8 -10   6   5  -4  -3   9  -1  -7   2]
     [  4   6   7  -1   9  -2  -3  10  -5  -8]
     [ -7   5  -8 -10  -2   9   1   3  -6   4]
     [-10   3  -2  -7  -6   5   4  -9   8   1]
     [  3  -4  -1   2  10  -8  -9   6   7  -5]
     [  9  -8  10  -5   4  -7   6   2  -1  -3]
     [ -3  -6   1   9   8   2  10  -5  -4  -7]
     [ -2   1   9   8 -10   7  -6  -4  -3   5]
     [  7   8 -10  -6  -9   4  -1  -2   5   3]
     [ -9   7  -4   3  -8 -10  -2   5   1   6]
     [  6  -5  -7  10   2  -1   3   9  -8  -4]]
    """
    y.fill(0)  # first zero the output matrix
    days, n = y.shape  # the number of days and teams to be scheduled
    div: Final[int] = n - 1  # the divisor for permutation values -> teams

    for game in x:
        g = int(game)  # make sure that we have the full integer range.
        home_idx: int = (g // div) % n  # home idx is in 0..n-1
        away_idx: int = g % div  # away index in 0..n-2
        if away_idx >= home_idx:  # "A vs. A" games impossible
            away_idx += 1  # away index in 0..n-1, but != home_idx

        for day in range(days):  # iterate over all possible rows for game
            if (y[day, home_idx] != 0) or (y[day, away_idx] != 0):
                continue  # day already blocked
            y[day, home_idx] = away_idx + 1
            y[day, away_idx] = -(home_idx + 1)
            break


class GameEncoding(Encoding):
    """An encoding that transforms strings of games to game plans."""

    def __init__(self, instance: Instance) -> None:
        """
        Create the game-based encoding.

        :param instance: the instance
        """
        super().__init__()
        #: the instance
        self.instance: Final[Instance] = instance
        self.decode = map_games  # type: ignore

    def search_space(self) -> Permutations:
        """
        Create a proper search space for this game-based encoding.

        The search space is a set of :mod:`~moptipy.spaces.permutations` that
        represents all the games that can take place in the tournament.
        Depending on the number of
        :attr:`~moptipyapps.ttp.instance.Instance.rounds` in the tournament,
        some games may appear multiple times. Home and away games are
        distributed in a fair and deterministic mannner between the teams.

        :return: the search space

        >>> inst = Instance.from_resource("circ4")
        >>> inst.n_cities
        4
        >>> inst.rounds
        2
        >>> ",".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0,1,2,3,4,5,6,7,8,9,10,11'
        >>> inst = Instance(inst.name, inst, inst.teams, inst.rounds,
        ...                 inst.home_streak_min, inst.home_streak_max,
        ...                 inst.away_streak_min, inst.away_streak_max,
        ...                 inst.separation_min, inst.separation_max)
        >>> inst.rounds = 1  # modify number of rounds for copied instance
        >>> ",".join(map(str, GameEncoding(inst).search_space().blueprint))
        '1,2,3,7,8,10'
        >>> inst.rounds = 3  # modify number of rounds for copied instance
        >>> ",".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0,1,1,2,2,3,3,4,5,6,7,7,8,8,9,10,10,11'
        >>> inst.rounds = 4  # modify number of rounds for copied instance
        >>> ",".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11'
        >>> inst.rounds = 5  # modify number of rounds for copied instance
        >>> ",".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0,0,1,1,1,2,2,2,3,3,3,4,4,5,5,6,6,7,7,7,8,8,8,9,9,10,10,10,11,11'
        """
        return search_space_for_n_and_rounds(
            self.instance.n_cities, self.instance.rounds)



r"""
A tool for collecting statistics.

>>> instance = Instance(
...     name="test2", n_products=2, n_customers=1, n_stations=2, n_demands=5,
...     time_end_warmup=21, time_end_measure=10000,
...     routes=[[0, 1], [1, 0]],
...     demands=[[0, 0, 1, 10, 20, 90], [1, 0, 0, 5, 22, 200],
...              [2, 0, 1, 7, 30, 200], [3, 0, 1, 6, 60, 200],
...              [4, 0, 0, 125, 5, 2000]],
...     warehous_at_t0=[2, 1],
...     station_product_unit_times=[[[10.0, 50.0, 15.0, 100.0],
...                                  [ 5.0, 20.0,  7.0,  35.0, 4.0, 50.0]],
...                                 [[ 5.0, 24.0,  7.0,  80.0],
...                                  [ 3.0, 21.0,  6.0,  50.0,]]])


>>> from moptipyapps.prodsched.simulation import Simulation
>>> statistics = Statistics(instance.n_products)
>>> collector = StatisticsCollector(instance)
>>> collector.set_dest(statistics)
>>> simulation = Simulation(instance, collector)
>>> simulation.ctrl_run()
>>> print("\n".join(str(statistics).split("\n")[:-1]))
stat;total;product_0;product_1
trp.mean;304.8888888888889;455.6;246.92307692307693
trp.sd;97.56326157594913;0;19.50721118346466
fill.rate;0;0;0
cwt.mean;1829.3333333333333;2278;1605
cwt.sd;388.56187838403986;nan;2.8284271247461903
stocklevel.mean;0.15813207736246118;0;0.15813207736246118
fulfilled.rate;1;1;1
"""

from time import time_ns
from typing import Final

from pycommons.math.stream_statistics import StreamStats, StreamSum
from pycommons.types import type_error

from moptipyapps.prodsched.instance import (
    Demand,
    Instance,
)
from moptipyapps.prodsched.simulation import Job, Listener
from moptipyapps.prodsched.statistics import Statistics


class StatisticsCollector(Listener):
    """A listener for simulation events."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the listener.

        :param instance: the instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the destination to write to.
        self.__dest: Statistics | None = None
        #: the end of the warmup period
        self.__warmup: Final[float] = instance.time_end_warmup
        #: the total time window length
        self.__total: Final[float] = instance.time_end_measure
        #: the number of products
        self.__n_products: Final[int] = instance.n_products
        #: the measurable demands per product
        self.__n_mdpb: Final[tuple[int, ...]] = (
            instance.n_measurable_demands_per_product)

        n_products: Final[int] = instance.n_products
        #: the internal per-product production time records
        self.__production_times: Final[tuple[StreamStats, ...]] = tuple(
            StreamStats() for _ in range(n_products))
        #: the internal total production time
        self.__production_time: Final[StreamStats] = StreamStats()
        #: the immediately satisfied products: satisfied vs. total
        self.__immediately_satisfied: Final[list[int]] = [0] * n_products
        #: the statistics of the waiting time for not-immediately satisfied
        #: products, per product
        self.__waiting_time_for_none_immediates: Final[tuple[
            StreamStats, ...]] = tuple(StreamStats() for _ in range(
                n_products))
        #: the total waiting time for non-immediately satisfied products
        self.__waiting_time_for_none_immediate: Final[StreamStats] = (
            StreamStats())
        #: the number of fulfilled jobs, per-product
        self.__fulfilled: Final[list[int]] = [0] * n_products
        #: the stock levels on a per-product basis
        self.__stock_levels: Final[tuple[StreamSum, ...]] = tuple(
            StreamSum() for _ in range(n_products))
        #: the total stock level sum
        self.__stock_level: Final[StreamSum] = StreamSum()
        #: the number of products currently in warehouse and since when
        #: they were there
        self.__in_warehouse: Final[list[list[float]]] = [[
            0.0, 0.0]] * n_products
        #: the stat time of the simulation
        self.__start: int = -1

    def set_dest(self, dest: Statistics) -> None:
        """
        Set the statistics record to fill.

        :param dest: the destination
        """
        if not isinstance(dest, Statistics):
            raise type_error(dest, "dest", Statistics)
        self.__dest = dest

    def start(self) -> None:
        """Clear all the data of the collector."""
        if self.__dest is None:
            raise ValueError("Need destination statistics!")
        self.__start = time_ns()
        n_products: Final[int] = self.__n_products
        for i in range(n_products):
            self.__production_times[i].reset()
            self.__immediately_satisfied[i] = 0
            self.__waiting_time_for_none_immediates[i].reset()
            self.__fulfilled[i] = 0
            self.__stock_levels[i].reset()
            wh = self.__in_warehouse[i]
            wh[0] = 0.0
            wh[1] = 0.0
        self.__production_time.reset()
        self.__waiting_time_for_none_immediate.reset()
        self.__stock_level.reset()

    def product_in_warehouse(
            self, time: float, product_id: int, amount: int,
            is_in_measure_period: bool) -> None:
        """
        Report a change of the amount of products in the warehouse.

        :param time: the current time
        :param product_id: the product ID
        :param amount: the new absolute total amount of that product in the
            warehouse
        :param is_in_measure_period: is this event inside the measurement
            period?
        """
        iwh: Final[list[float]] = self.__in_warehouse[product_id]
        if is_in_measure_period:
            value: float = (time - max(iwh[1], self.__warmup)) * iwh[0]
            self.__stock_levels[product_id].add(value)
            self.__stock_level.add(value)
        iwh[0] = amount
        iwh[1] = time

    def produce_at_end(
            self, time: float, station_id: int, job: Job) -> None:
        """
        Report the completion of the production of a product at a station.

        :param time: the current time
        :param station_id: the station ID
        :param job: the production job
        """
        if job.measure and job.completed:
            am: Final[int] = job.amount
            tt: float = time - job.arrival
            if am <= 1:
                self.__production_times[job.product_id].add(tt)
                self.__production_time.add(tt)
            else:  # deal with amounts > 1
                tt /= am
                for _ in range(am):
                    self.__production_times[job.product_id].add(tt)
                    self.__production_time.add(tt)

    def demand_satisfied(
            self, time: float, demand: Demand) -> None:
        """
        Report that a given demand has been satisfied.

        :param time: the time index when the demand was satisfied
        :param demand: the demand that was satisfied
        :param is_in_measure_period: is this event inside the measurement
            period?
        """
        if demand.measure:
            at: Final[float] = demand.arrival
            pid: Final[int] = demand.product_id
            if at >= time:
                self.__immediately_satisfied[pid] += 1
            else:
                tt: Final[float] = time - at
                self.__waiting_time_for_none_immediates[pid].add(tt)
                self.__waiting_time_for_none_immediate.add(tt)
            self.__fulfilled[pid] += 1

    def finished(self, time: float) -> None:
        """
        Fill the collected statistics into the statistics record.

        :param time: the time when we are finished
        """
        dest: Final[Statistics | None] = self.__dest
        if dest is None:
            raise ValueError("Lost destination statistics record?")
        total: Final[tuple[int, ...]] = self.__n_mdpb

        for i, stat in enumerate(self.__production_times):
            dest.production_time_means[i] = stat.mean()
            dest.production_time_sds[i] = stat.sd()
        dest.production_time_mean = self.__production_time.mean()
        dest.production_time_sd = self.__production_time.sd()

        sn: int = 0
        sf: int = 0
        st: int = 0
        for i, n in enumerate(self.__immediately_satisfied):
            t = total[i]
            dest.immediately_satisfied_rates[i] = n / t
            sn += n
            f = self.__fulfilled[i]
            dest.fulfilled_rates[i] = f / t
            sf += f
            st += t
        dest.immediately_satisfied_rate = sn / st
        dest.fulfilled_rate = sf / st

        for i, stat in enumerate(self.__waiting_time_for_none_immediates):
            dest.waiting_time_for_none_immediates_means[i] = stat.mean()
            dest.waiting_time_for_none_immediates_sds[i] = stat.sd()
        dest.waiting_time_for_none_immediates_mean = (
            self.__waiting_time_for_none_immediate.mean())
        dest.waiting_time_for_none_immediates_sd = (
            self.__waiting_time_for_none_immediate.sd())

        slm: Final[StreamSum] = self.__stock_level
        twl: Final[float] = self.__total - self.__warmup
        for i, sm in enumerate(self.__stock_levels):
            wh = self.__in_warehouse[i]
            v: float = (self.__total - wh[1]) * wh[0]
            sm.add(v)
            slm.add(v)
            dest.stock_level_means[i] = sm.result() / twl
        dest.stock_level_mean = slm.result() / twl
        dest.simulation_time_nanos = time_ns() - self.__start



"""
A permutation-with-repetition-based encoding based on games with rescheduling.

This game-based encoding method is an extension of the method presented in
:mod:`~moptipyapps.ttp.game_encoding`. The method presented in
:mod:`~moptipyapps.ttp.game_encoding` takes a permutation of games and fills
these games into a game plan one-by-one. For each game it tries to find the
earliest possible slot where it can be scheduled, i.e., the earliest slot
where both involved teams do not yet have a game scheduled.

During this process, it can happen that a game cannot be added to the game
plan, because on each day, at least one of the two teams has already a game
scheduled. In such a case, the original encoding will simply drop the game.
This results in "byes" in the game plan.

This new encoding tries to reduce the number of such byes. It does this as
follows: It still processes the game permutations from front to end and
adds the games to the schedule. If one game cannot be added to the schedule
due to the above reason, it will go through game plan again from beginning
to end. This time, it will look for a day on which only *one* of the two teams
has a game scheduled already. It will then take this already-scheduled game
and try to schedule it on a *later* day. This is a recursive process: If no
later slot is found where both teams involved have no game scheduled, we will
again look for a day on which only one of the teams have scheduled a game and
try to move that to a later slot. Since this will be an even later slot, the
recursion will never go too deep. Either way, if a later slot is found, the
game can be moved, meaning that the original game blocking our "current" game
can be moved as well. Then the new game can be inserted.

If no such slot can be found, then we try to insert the game by searching
from the back end and moving other games forward in the same recursive
fashion. The hope is that, this way, the number of "byes" in the schedule can
be reduced. If the number of "byes" is reduced, then there will automatically
fewer constraint violations. This makes it more likely that we discover a
feasible schedule earlier in the search.

Furthermore, the process might even make it easier for a search to move from
one feasible schedule to another one. There would hopefully be fewer
permutations leading to infeasible schedules and fewer infeasible schedules
between two feasible ones.

Of course, all of that comes at the cost of increased runtime.
"""


from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.ttp.game_encoding import GameEncoding
from moptipyapps.ttp.game_plan_space import GamePlanSpace
from moptipyapps.ttp.instance import Instance


# Due to the recursion, caching must be disabled or we get a segfault.
@numba.njit(cache=False, nogil=True, fastmath=True, boundscheck=True)
def _re_schedule(y: np.ndarray, index: int | np.ndarray,
                 day: int, end: int, direction: int) -> bool:
    """
    Try to re-schedule a game at the given index.

    This function tries to take the game at the given index and given day and
    attempts to move it to a later day. If it succeeds, it returns `True`, if
    rescheduling was not possible, it returns `False`.

    :param y: the schedule
    :param index: the index
    :param day: the day
    :param end: the end day
    :return: `True` if the game was successfully re-scheduled, `False`
        otherwise

    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 4,  3, -2, -1],
    ...     [-2,  1, -4,  3],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _re_schedule(dest, 0, 2, 6, 1)
    False
    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 4,  3, -2, -1],
    ...     [ 0,  1, -4,  0],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _re_schedule(dest, 0, 2, 6, 1)
    True
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 0  3 -2  0]
     [ 4  1 -4 -1]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 0,  3, -2, -1],
    ...     [-2,  0,  0,  3],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _re_schedule(dest, 2, 1, 6, 1)
    True
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 0  4  0 -2]
     [ 3  0 -1 -1]
     [-2  3 -2  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    """
    home_idx = int(index)  # We first guess that this is the home team
    away_idx = int(y[day, index])  # and that this is the away team.
    if away_idx < 0:  # The home team actually plays away.
        away_idx, home_idx = home_idx, (-away_idx) - 1  # Swap.
    else:  # We were right.
        away_idx -= 1  # So the ID of the away team -1 is its index.
    if _schedule(y, home_idx, away_idx, day + direction,
                 end, direction, True):  # Try to re-schedule the game.
        y[day, home_idx] = 0  # Set first slot to empty.
        y[day, away_idx] = 0  # Set last slot to empty.
        return True  # We successfully rescheduled the game.
    return False  # We failed to recursively reschedule.


# Due to the recursion, caching must be disabled or we get a segfault.
@numba.njit(cache=False, nogil=True, fastmath=True, boundscheck=True)
def _schedule(y: np.ndarray, home_idx: int, away_idx: int, start: int,
              end: int, direction: int, is_recursive: bool) -> bool:
    """
    Try to find a slot in the day range where the given game can be placed.

    :param y: the destination array
    :param home_idx: the home index
    :param away_idx: the away index
    :param start: the starting day
    :param end: the ending day
    :return: `True` if the game could be placed, `False` otherwise.

    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 4,  3, -2, -1],
    ...     [-2,  1, -4,  3],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _schedule(dest, 1, 2, 0, 6, 1, True)
    False
    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 4,  0,  0, -1],
    ...     [-2,  1, -4,  3],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _schedule(dest, 1, 2, 0, 6, 1, True)
    True
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 4  3 -2 -1]
     [-2  1 -4  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    >>> dest = np.array([
    ...     [ 2, -1,  4, -3],
    ...     [ 3,  4, -1, -2],
    ...     [ 0,  3, -2, -1],
    ...     [-2,  0,  0,  3],
    ...     [-3, -4,  1,  2],
    ...     [-4, -3,  2,  1]])
    >>> _schedule(dest, 0, 2, 0, 6, 1, True)
    True
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 3  0 -1 -1]
     [-2  3 -2  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    """
    empty_slots_needed = 2  # Number of required open slots for a game.
    while True:
        for day in range(start, end, direction):  # iterate over all days
            can_use_home: bool = y[day, home_idx] == 0  # Check for empty slot
            can_use_away: bool = y[day, away_idx] == 0  # Check for empty slot
            if can_use_home + can_use_away < empty_slots_needed:
                continue  # Too few empty slots: Try with next day.
            if ((can_use_home or _re_schedule(  # At least one empty slot?
                    y, home_idx, day, end, direction)) and (
                    can_use_away or _re_schedule(
                    y, away_idx, day, end, direction))):
                # There either were two empty slots, or one and the other game
                # was rescheduled. So we can schedule the game.
                y[day, home_idx] = away_idx + 1
                y[day, away_idx] = -(home_idx + 1)
                return True  # Success!

        # After we tried finding a perfect slot, we now try with just one slot.
        if empty_slots_needed >= 2:
            empty_slots_needed = 1  # Try requiring just one slot.
            continue  # And execute scheduling loop again.
        if is_recursive:  # Only if this is the root call, we can go one
            return False  # after the attempt requiring 1 slot failed.
        is_recursive = True  # But we can only do this once.
        # If we got here, we are in the root call.
        # We first tried to find two open slots and failed.
        # Then we tried to find a single open slot and tried to re-schedule
        # the other game to a later point in time ... an failed.
        # Now we try rescheduling games towards an early time.
        start = end - 1
        end = direction = -1


# Due to the recursion, caching must be disabled or we get a segfault.
@numba.njit(cache=False, fastmath=True, boundscheck=True)
def map_games_2(x: np.ndarray, y: np.ndarray) -> None:
    """
    Translate a permutation of games to a game plan.

    This is a straightforward decoding that places the games into the map one
    by one. Each game is placed at the earliest slot in which it can be
    placed. If a game cannot be placed, it is ignored. This will lead to many
    errors, which can be counted via the :mod:`~moptipyapps.ttp.errors`
    objective.

    :param x: the source permutation
    :param y: the destination game plan

    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> from moptipyapps.ttp.game_encoding import (
    ...     search_space_for_n_and_rounds)
    >>> teams = 2
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ 2 -1]
     [-2  1]]
    >>> teams = 4
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 4  3 -2 -1]
     [-2  1 -4  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
     >>> from moptipyapps.ttp.instance import Instance
     >>> inst = Instance.from_resource("circ10")
     >>> perm = np.array([73,77,55,74,21,20,3,11,63,19,38,8,27,47,88,16,75,
     ...     45,89,36,24,80,17,40,0,32,53,82,31,13,12,66,71,6,87,84,2,60,61,
     ...     10,30,58,49,57,54,23,26,46,59,42,29,62,22,18,50,78,37,34,65,43,
     ...     48,85,86,83,56,41,5,81,52,15,51,9,4,76,7,69,67,33,79,72,35,25,
     ...     1,14,70,44,68,64,28,39])
    >>> teams = 10
    >>> rounds = 2
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ -8  -9   5   7  -3  10  -4   1   2  -6]
     [  5  -7   4  -3  -1  -9   2 -10   6   8]
     [ 10   4  -9  -2   6  -5   8  -7   3  -1]
     [ -4  -3   2   1  -7   8   5  -6 -10   9]
     [ -6   9  -5  -8   3   1 -10   4  -2   7]
     [ -5  10  -6  -9   1   3  -8   7   4  -2]
     [  2  -1   8   6   7  -4  -5  -3  10  -9]
     [  8 -10   6   5  -4  -3   9  -1  -7   2]
     [  4   6   7  -1   9  -2  -3  10  -5  -8]
     [ -7   5  -8 -10  -2   9   1   3  -6   4]
     [-10   3  -2  -7  -6   5   4  -9   8   1]
     [  3  -4  -1   2  10  -8  -9   6   7  -5]
     [  9  -8  10  -5   4  -7   6   2  -1  -3]
     [ -3  -6   1   9   8   2  10  -5  -4  -7]
     [ -2   1   9   8 -10   7  -6  -4  -3   5]
     [  7   8 -10  -6  -9   4  -1  -2   5   3]
     [ -9   7  -4   3  -8 -10  -2   5   1   6]
     [  6  -5  -7  10   2  -1   3   9  -8  -4]]
    >>> int(dest.shape[0] * dest.shape[1] - np.count_nonzero(dest))
    0
    >>> from moptipyapps.ttp.game_encoding import re_encode
    >>> re_encode(perm, dest)
    >>> print(perm)
    [21 32 53 63 73  3 20 55 77 88  8 11 40 60 74 19 27 51 58 89 16 38 45 66
     87 17 36 47 69 75  0 24 31 41 80  6 22 30 61 82  2 13 23 43 71 12 52 54
     65 84 10 49 57 79 81  1 28 44 68 78  7 26 39 59 64 18 34 42 46 62  9 25
     33 50 85  5 15 48 76 83 14 29 67 72 86  4 35 37 56 70]
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ -8  -9   5   7  -3  10  -4   1   2  -6]
     [  5  -7   4  -3  -1  -9   2 -10   6   8]
     [ 10   4  -9  -2   6  -5   8  -7   3  -1]
     [ -4  -3   2   1  -7   8   5  -6 -10   9]
     [ -6   9  -5  -8   3   1 -10   4  -2   7]
     [ -5  10  -6  -9   1   3  -8   7   4  -2]
     [  2  -1   8   6   7  -4  -5  -3  10  -9]
     [  8 -10   6   5  -4  -3   9  -1  -7   2]
     [  4   6   7  -1   9  -2  -3  10  -5  -8]
     [ -7   5  -8 -10  -2   9   1   3  -6   4]
     [-10   3  -2  -7  -6   5   4  -9   8   1]
     [  3  -4  -1   2  10  -8  -9   6   7  -5]
     [  9  -8  10  -5   4  -7   6   2  -1  -3]
     [ -3  -6   1   9   8   2  10  -5  -4  -7]
     [ -2   1   9   8 -10   7  -6  -4  -3   5]
     [  7   8 -10  -6  -9   4  -1  -2   5   3]
     [ -9   7  -4   3  -8 -10  -2   5   1   6]
     [  6  -5  -7  10   2  -1   3   9  -8  -4]]
     >>> inst = Instance.from_resource("con14")
     >>> perm = np.array([70,114,54,51,98,31,49,46,148,169,174,151,155,110,
     ...     75,118,128,81,171,19,133,93,5,91,47,1,84,109,142,121,7,40,163,
     ...     61,20,96,165,177,160,39,73,37,101,108,119,65,2,117,164,106,60,
     ...     145,105,90,170,74,8,26,94,10,22,86,120,154,89,66,77,178,140,48,
     ...     0,52,159,25,176,135,63,57,76,161,68,124,162,29,55,80,24,45,85,
     ...     115,131,100,17,53,27,38,95,158,104,175,87,123,167,97,79,112,
     ...     172,127,136,12,44,50,102,69,144,143,113,4,181,88,166,150,59,
     ...     141,72,125,116,132,14,147,153,129,111,92,62,6,32,82,16,179,21,
     ...     3,126,134,122,149,146,36,107,34,103,42,41,18,35,78,28,137,43,
     ...     33,71,99,139,56,23,138,67,168,130,30,180,152,83,9,13,173,11,
     ...     157,156,64,15,58])
    >>> teams = 14
    >>> rounds = 2
    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> from moptipyapps.ttp.game_encoding_2 import map_games_2
    >>> map_games_2(perm, dest)
    >>> print(dest)
    [[ -8 -10  -5  14   3   7  -6   1  12   2 -13  -9  11  -4]
     [-14  -6   7  12  11   2  -3   9  -8  13  -5  -4 -10   1]
     [  7   8 -14   9 -10 -12  -1  -2  -4   5  13   6 -11   3]
     [ -5  11  -8  -7   1 -14   4   3 -12 -13  -2   9  10   6]
     [  3  -5  -1 -11   2  10  -9 -13   7  -6   4  14   8 -12]
     [  9  -3   2  10 -13  12   8  -7  -1  -4  14  -6   5 -11]
     [-10  -4  13   2 -11  -9  14  12   6   1   5  -8  -3  -7]
     [ -4   9 -10   1   7  -8  -5   6  -2   3 -14  13 -12  11]
     [ -6 -11 -12  -8  10   1  13   4 -14  -5   2   3  -7   9]
     [  4 -14 -13  -1  -9  11  10 -12   5  -7  -6   8   3   2]
     [ 10  -7   5   8  -3  14   2  -4 -13  -1  12 -11   9  -6]
     [ 12  14  -9 -10  13 -11  -8   7   3   4   6  -1  -5  -2]
     [ -3  -9   1  11  -8  13  12   5   2 -14  -4  -7  -6  10]
     [  2  -1  -7 -13  -6   5   3 -14  10  -9 -12  11   4   8]
     [-11 -12  14  -5   4 -13   9 -10  -7   8   1   2   6  -3]
     [ -9   3  -2  -6 -14   4 -13  11   1  12  -8 -10   7   5]
     [-12  13   8   5  -4 -10 -14  -3  11   6  -9   1  -2   7]
     [  8   6  10 -14 -12  -2  11  -1  13  -3  -7   5  -9   4]
     [ 14  -8 -11   6   9  -4 -10   2  -5   7   3 -13  12  -1]
     [ -2   1  -6  13  12   3 -11  14 -10   9   7  -5  -4  -8]
     [ 11   5  12   7  -2   9  -4  13  -6  14  -1  -3  -8 -10]
     [  6  10  11 -12  -7  -1   5  -9   8  -2  -3   4 -14  13]
     [  5 -13  -4   3  -1   8 -12  -6  14  11 -10   7   2  -9]
     [-13   7   6  -9  14  -3  -2 -11   4 -12   8  10   1  -5]
     [ -7  12   4  -3   6  -5   1  10 -11  -8   9  -2  14 -13]
     [ 13   4   9  -2   8  -7   6  -5  -3 -11  10 -14  -1  12]]
    """
    y.fill(0)  # first zero the output matrix
    days, n = y.shape  # the number of days and teams to be scheduled
    div: Final[int] = n - 1  # the divisor for permutation values -> teams

    for game in x:
        g = int(game)  # make sure that the indices are Python integers
        home_idx: int = (g // div) % n  # home idx is in 0..n-1
        away_idx: int = g % div  # away index in 0..n-2
        if away_idx >= home_idx:  # "A vs. A" games impossible
            away_idx += 1  # away index in 0..n-1, but != home_idx
        _schedule(y, home_idx, away_idx, 0, days, 1, False)


class GameEncoding2(GameEncoding):
    """A second encoding that transforms strings of games to game plans."""

    def __init__(self, instance: Instance) -> None:
        """
        Create the game-based encoding that can reschedule games.

        :param instance: the instance
        """
        super().__init__(instance)
        self.decode = map_games_2  # type: ignore

        # We invoke all the functions in order to enforce proper compilation.
        # Numba has some problems with recursive functions.
        # Therefore, we try to enforce that all functions used here are
        # invoked in an order where their parameters can be fully inferred.
        y: np.ndarray = GamePlanSpace(instance).create()
        y.fill(1)
        _re_schedule(y, 0, 0, 1, 1)
        _schedule(y, 0, 1, 0, 1, 1, False)
        y[0, 0] = 0
        _schedule(y, 0, 1, 0, 1, 1, True)
        y.fill(0)
        _re_schedule(y, 0, 0, 1, 1)
        _schedule(y, 0, 1, 0, 1, 1, False)
        x: np.ndarray = self.search_space().blueprint
        x.sort()
        map_games_2(x, y)
